#!/usr/bin/env python3

from collections import OrderedDict
from filelock import FileLock
import importlib
import os
import paho.mqtt.client as mqtt
import RPi.GPIO as GPIO
from serial import Serial
from smbus2 import SMBus
import socket
import sys
import time
import toml
import traceback


ACT_PIN_ID = "ACTIVATION_PIN"
LOCK_PREFIX = "/run/lock/piot"


def get_lock(lock_file):
    os.makedirs(os.path.dirname(lock_file), exist_ok=True)
    if not os.path.isfile(lock_file):
        os.mknod(lock_file)
    return FileLock(lock_file)


def find(obj, key):
    if isinstance(obj, dict):
        if key in obj:
            yield obj[key]
        for v in obj.values():
            yield from find(v, key)
    elif isinstance(obj, list):
        for v in obj:
            yield from find(v, key)


def format_msg(timestamp, measurement, tags, fields):
    tstr = ",".join(["{}={}".format(k, v) for k, v in tags.items()])
    fstr = ",".join(["{}={}".format(k, v) for k, v in fields.items()])
    return "{},{} {} {}".format(measurement, tstr, fstr, timestamp)


def sync_wait(sync):
    return sync - time.time() % sync if sync > 0 else 0


def round_step(x, step):
    return x // step * step if step else x


def run_drivers(cfg, sync=0):
    time.sleep(sync_wait(sync))
    for driver_id in cfg:
        try:
            for dcfg in cfg[driver_id]:
                activation_pin = None
                if ACT_PIN_ID in dcfg:
                    activation_pin = dcfg[ACT_PIN_ID]
                    dcfg = {k: v for k, v in dcfg.items() if k != ACT_PIN_ID}
                driver_module = importlib.import_module("piot.drivers." + driver_id)
                with ActivationContext(activation_pin), \
                     getattr(driver_module, "Driver")(**dcfg) as driver:
                    res = driver.run()
                    if not res:
                        continue
                    for did, ts, fields in res:
                        yield did, round_step(ts, sync), fields
        except (KeyboardInterrupt, SystemExit):
            raise
        except:
            traceback.print_exc()


def run(cfg, measurement, host, client=None, qos=0, sync=0):
    for driver_id, ts, fields in run_drivers(cfg, sync):
        if fields:
            fields = OrderedDict([(k, v) \
                for k, v in fields.items() if v is not None])
        if not fields:
            continue
        tags = OrderedDict([
            ("host", host),
            ("sid", driver_id),
        ])
        payload = format_msg(ts, measurement, tags, fields)
        if client:
            topic = "{}/{}/{}".format(measurement, host, driver_id)
            client.publish(topic, payload, qos, True)
        else:
            print(payload)


def main():
    if len(sys.argv) <= 1:
        print("Usage: {} <cfg_file>".format(sys.argv[0]))
    cfg_file  = sys.argv[1]

    # Read configuration
    cfg = toml.load(cfg_file)
    interval = cfg.get("interval", 0)
    measurement = cfg.get("measurement", "data")
    host = cfg.get("host", socket.gethostname())
    drivers_cfg = cfg.get("drivers", {})

    # Connect to MQTT broker, if necessary
    mqtt_client, mqtt_qos = None, 0
    mqtt_cfg = cfg.get("mqtt", None)
    if mqtt_cfg is not None:
        mqtt_host = mqtt_cfg.get("host", "localhost")
        mqtt_port = mqtt_cfg.get("port", 1883)
        mqtt_qos = mqtt_cfg.get("qos", 2)
        print(f"Connecting to MQTT broker at '{mqtt_host}:{mqtt_port}'")
        mqtt_client = mqtt.Client(host, clean_session=False)
        mqtt_client.connect(mqtt_host, mqtt_port)
        mqtt_client.loop_start()

    # Run drivers
    try:
        with GPIOContext(drivers_cfg):
            while True:
                run(drivers_cfg, measurement, host,
                    mqtt_client, mqtt_qos, interval)
    finally:
        if mqtt_client:
            mqtt_client.disconnect()


class GPIOContext:
    def __init__(self, cfg):
        GPIO.setmode(GPIO.BCM)
        pin_list = list(find(cfg, ACT_PIN_ID))
        GPIO.setup(pin_list, GPIO.OUT, initial=GPIO.HIGH)


    def close(self):
        GPIO.cleanup()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class ActivationContext:
    def __init__(self, pin=None):
        self._open = False
        self._pin = pin
        if self._pin:
            self._lock = get_lock(
                "{}/gpio{}.lock".format(LOCK_PREFIX, self._pin))


    def open(self):
        if not self._pin or self._open:
            return
        self._open = True
        if self._lock: self._lock.acquire()
        GPIO.output(self._pin, GPIO.LOW)


    def close(self):
        if not self._pin or not self._open:
            return
        GPIO.output(self._pin, GPIO.HIGH)
        if self._lock: self._lock.release()
        self._open = False


    def __enter__(self):
        self.open()
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class DriverBase:
    def __init__(self, lock_file=None):
        self._open = True
        self._lock = get_lock(lock_file) if lock_file else None
        if self._lock: self._lock.acquire()


    def close(self):
        if not self._open:
            return
        if self._lock: self._lock.release()
        self._open = False


    def sid(self):
        return self.__class__.__module__.split(".")[-1]


    def run(self):
        raise NotImplementedError()


    def __enter__(self):
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        self.close()


class I2CDriver(DriverBase):
    def __init__(self):
        super().__init__("{}/i2c.lock".format(LOCK_PREFIX))


class SMBusDriver(I2CDriver):
    def __init__(self, bus=1):
        super().__init__()
        self._bus = SMBus(bus)


    def close(self):
        if not self._open:
            return
        if self._bus: self._bus.close()
        super().close()


class SerialDriver(DriverBase):
    def __init__(self, *args, **kwargs):
        super().__init__("{}/serial.lock".format(LOCK_PREFIX))
        self._serial = Serial(timeout=1, *args, **kwargs)
        self._serial.reset_input_buffer()
        self._serial.reset_output_buffer()


    def close(self):
        if not self._open:
            return
        if self._serial: self._serial.close()
        super().close()


    def _cmd(self, cmd, size=1):
        self._serial.write(cmd)
        self._serial.flush()
        time.sleep(0.1)
        return self._serial.read(size)


if __name__ == "__main__":
    main()



from collections import OrderedDict
import time

from ..core import I2CDriver
import board
import busio
from adafruit_bme280 import Adafruit_BME280_I2C


class Driver(I2CDriver):
    def __init__(self, address=0x77):
        super().__init__()
        i2c = busio.I2C(board.SCL, board.SDA)
        self._sensor = Adafruit_BME280_I2C(i2c, address=address)


    def run(self):
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("temperature", self._sensor.temperature),
            ("humidity",    self._sensor.humidity),
            ("pressure",    self._sensor.pressure),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SMBusDriver


_REG_DATA        = 0x00
_REG_STATUS      = 0x06
_REG_TEMPERATURE = 0x07
_REG_CTRL1       = 0x09
_REG_CTRL2       = 0x0A
_REG_PERIOD      = 0x0B


class Driver(SMBusDriver):
    def __init__(self, address=0x0D, bus=1):
        super().__init__(bus)
        self._address = address
        self._bus.write_byte_data(self._address, _REG_CTRL1, 0x01)


    def run(self):
        ts = int(time.time() * 1e9)
        res = self._bus.read_i2c_block_data(self._address, _REG_DATA, 9)
        x, y, z, status, temperature = struct.unpack("<hhhBh", bytearray(res))
        return [(self.sid(), ts, OrderedDict([
            ("mag_x", x),
            ("mag_y", y),
            ("mag_z", z),
            ("status", status),
            ("temperature", temperature),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SMBusDriver


_REG_CFGA = 0x00
_REG_CFGB = 0x01
_REG_MODE = 0x02
_REG_DATA = 0x03


class Driver(SMBusDriver):
    def __init__(self, address=0x1E, bus=1):
        super().__init__(bus)
        self._address = address
        self._bus.write_byte_data(self._address, _REG_MODE, 0x00)


    def run(self):
        ts = int(time.time() * 1e9)
        res = self._bus.read_i2c_block_data(self._address, _REG_DATA, 7)
        x, y, z, status = struct.unpack(">hhhB", bytearray(res))
        return [(self.sid(), ts, OrderedDict([
            ("mag_x", x),
            ("mag_y", y),
            ("mag_z", z),
            ("status", status),
        ]))]



from collections import OrderedDict
import struct
import sys
import time

from ..core import SMBusDriver, run_drivers
from smbus2 import SMBus


class Driver(SMBusDriver):
    _CMD_MOVE = 0x4D
    _CMD_READ = 0x52


    def __init__(self, address, bus=1, movement=None, drivers=None,
        interval=0, polling_interval=0.1, check_move=True):
        super().__init__(bus)
        self._address = address
        self._busnum = bus
        self._movement = movement
        self._drivers = drivers
        self._interval = interval
        self._polling_interval = polling_interval
        self._check_move = check_move


    def run(self):
        self._move(0, 0, 0)
        for vert in _get_range(self._movement["vert"]):
            for pan in _get_range(self._movement["pan"]):
                for tilt in _get_range(self._movement["tilt"]):
                    self._move(vert, pan, tilt)
                    vert, pan, tilt, flags, bt1, bt2 = self._read_state()
                    state = OrderedDict([
                        ("vert"    , vert),
                        ("pan"     , pan),
                        ("tilt"    , tilt),
                        ("flags"   , flags),
                        ("battery1", bt1),
                        ("battery2", bt2),
                    ])
                    yield self.sid(), int(time.time() * 1e9), state
                    if self._drivers:
                        self._bus.close()
                        if self._lock: self._lock.release()
                        yield from run_drivers(self._drivers, self._interval)
                        if self._lock: self._lock.acquire()
                        self._bus = SMBus(self._busnum)


    def _move(self, vert, pan, tilt):
        data = struct.pack(">HBB", vert, pan, tilt)
        _retry(lambda: self._bus.write_i2c_block_data(
            self._address, self._CMD_MOVE, data), 0.5)
        if self._check_move:
            while True:
                time.sleep(self._polling_interval)
                cvert, cpan, ctilt, _, _, _ = self._read_state()
                if cvert == vert and cpan == pan and ctilt == tilt:
                    break
        else:
            time.sleep(self._polling_interval)
        time.sleep(0.1)


    def _read_state(self):
        data = _retry(lambda: self._bus.read_i2c_block_data(
            self._address, self._CMD_READ_STATE, 7), 0.5)
        return struct.unpack(">HBBBBB", data)


def _retry(func, interval):
    while True:
        try:
            return func()
        except OSError:
            print("[vertpantilt] OSError", file=sys.stderr)
            time.sleep(interval)


def _get_range(cfg):
    return range(cfg["start"], cfg["stop"] + 1, cfg["step"])


def _close(a, b, tolerance):
    return abs(a - b) <= tolerance



from collections import OrderedDict
import time

from ..core import I2CDriver
import board
import busio
from adafruit_adxl34x import ADXL345


class Driver(I2CDriver):
    def __init__(self, address=0x53):
        super().__init__()
        i2c = busio.I2C(board.SCL, board.SDA)
        self._sensor = ADXL345(i2c, address=address)


    def run(self):
        ts, x, y, z = int(time.time() * 1e9), *self._sensor.acceleration
        return [(self.sid(), ts, OrderedDict([
            ("accel_x", x),
            ("accel_y", y),
            ("accel_z", z),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SerialDriver
from serial import SerialException


_PASSIVE_CODES = b"\x02\x01\x01"
_REQUEST_CODES = b"\x04\x00\x00"
_SETWORK_CODES = b"\x06\x01\x01"
_SETCONT_CODES = b"\x08\x01\x00"


def _seq(codes, device_id=b"\xFF\xFF"):
    head = b"\xAA\xB4" + codes + b"\x00" * 10 + device_id
    return head + bytes([_checksum(head)]) + b"\xAB"


def _checksum(res):
    return sum(res[2:]) & 0xFF


def _check(res):
    return res[-1:] != b"\xAB" or _checksum(res[:-2]) != res[-2]


class Driver(SerialDriver):
    def __init__(self, port):
        super().__init__(port, 9600)
        res = self._cmd(_seq(_PASSIVE_CODES), 10)
        if res[0:3] != b"\xAA\xC5\x02" or _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        res = self._cmd(_seq(_SETWORK_CODES), 10)
        if res[0:3] != b"\xAA\xC5\x06" or _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        res = self._cmd(_seq(_SETCONT_CODES), 10)
        if res[0:3] != b"\xAA\xC5\x08" or _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        

    def run(self):
        ts = int(time.time() * 1e9)
        res = self._cmd(_seq(_REQUEST_CODES), 10)
        if res[0:2] != b"\xAA\xC0" or _check(res):
            return ts, None
        pm25, pm10 = [x / 10 for x in struct.unpack("<HH", res[2:6])]
        return [(self.sid(), ts, OrderedDict([
            ("pm25", pm25),
            ("pm10", pm10),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SMBusDriver


_REG_WHO_AM_I   = 0x00
_REG_SMPLRT_DIV = 0x15
_REG_DLPF_FS    = 0x16
_REG_INT_CFG    = 0x17
_REG_INT_STATUS = 0x1A
_REG_DATA       = 0x1B
_REG_PWR_MGM    = 0x3E


class Driver(SMBusDriver):
    def __init__(self, address=0x68, bus=1):
        super().__init__(bus)
        self._address = address


    def run(self):
        ts = int(time.time() * 1e9)
        res = self._bus.read_i2c_block_data(self._address, _REG_DATA, 8)
        temperature, x, y, z = struct.unpack(">hhhh", bytearray(res))
        return [(self.sid(), ts, OrderedDict([
            ("temperature", temperature),
            ("gyro_x", x),
            ("gyro_y", y),
            ("gyro_z", z),
        ]))]



from collections import OrderedDict
import time

from ..core import I2CDriver
import board
import busio
from adafruit_amg88xx import AMG88XX


class Driver(I2CDriver):
    def __init__(self, address=0x69):
        super().__init__()
        i2c = busio.I2C(board.SCL, board.SDA)
        self._sensor = AMG88XX(i2c, addr=address)


    def run(self):
        ts, temp, px = int(time.time() * 1e9), \
            self._sensor.temperature, self._sensor.pixels
        data = OrderedDict()
        data["temperature"] = temp
        for j, row in enumerate(px):
            for i, val in enumerate(row):
                data["px{}{}".format(i, j)] = val
        return [(self.sid(), ts, data)]



from collections import OrderedDict
import time

from ..core import DriverBase
from adafruit_hcsr04 import HCSR04


class Driver(DriverBase):
    def __init__(self, trigger_pin, echo_pin):
        super().__init__()
        self._sensor = HCSR04(trigger_pin, echo_pin)


    def run(self):
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("distance", self._sensor.distance),
        ]))]


    def close(self):
        if self._sensor: self._sensor.deinit()
        super().close()


    def __enter__(self):
        super().__enter__()
        self._sensor.__enter__()
        return self


    def __exit__(self, exc_type, exc_value, traceback):
        self.close()



from collections import OrderedDict
import time

from ..core import I2CDriver
import board
import busio
from adafruit_tsl2561 import TSL2561


class Driver(I2CDriver):
    def __init__(self, address=0x39, gain=None, integration_time=None):
        super().__init__()
        i2c = busio.I2C(board.SCL, board.SDA)
        self._sensor = TSL2561(i2c, address)
        if gain:
            self._sensor.gain = gain
        if integration_time:
            self._sensor.integration_time = integration_time


    def run(self):
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("lux", self._sensor.lux),
            ("broadband", self._sensor.broadband),
            ("infrared", self._sensor.infrared),
        ]))]



from collections import OrderedDict
import time

from ..core import DriverBase
from adafruit_dht import DHT11, DHT22


class Driver(DriverBase):
    def __init__(self, pin, dht11):
        super().__init__()
        self._sensor = DHT11(pin) if dht11 else DHT22(pin)


    def run(self):
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("temperature", self._sensor.temperature),
            ("humidity",    self._sensor.humidity),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SerialDriver
from serial import SerialException


_REQUEST_SEQ = b"\xFF\x01\x86\x00\x00\x00\x00\x00\x79"
_CALZERO_SEQ = b"\xFF\x01\x87\x00\x00\x00\x00\x00\x78"
_CALSPAN_SEQ = b"\xFF\x01\x88\x07\xD0\x00\x00\x00\xA0"


def _checksum(res):
    return 0xFF - (sum(res[1:]) & 0xFF) + 1


def _check(res):
    return _checksum(res[:-1]) != res[-1]


class Driver(SerialDriver):
    def __init__(self, port):
        super().__init__(port, 9600)


    def run(self):
        ts = int(time.time() * 1e9)
        res = self._cmd(_REQUEST_SEQ, 9)
        if res[0:2] != b"\xFF\x86" or _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        co2, = struct.unpack(">H", res[2:4])
        return [(self.sid(), ts, OrderedDict([
            ("co2", co2),
        ]))]



from collections import OrderedDict
import datetime
import socket
import subprocess
import time

from ..core import DriverBase
from PIL import Image, ImageDraw, ImageFont
from Adafruit_SSD1306 import SSD1306_128_32, SSD1306_128_64



class Driver(DriverBase):
    def __init__(self, model64, rst=None):
        super().__init__()
        self._disp = SSD1306_128_64(rst) if model64 else SSD1306_128_32(rst)


    def run(self):
        # Initialize library.
        self._disp.begin()
        
        # Clear display.
        self._disp.clear()
        self._disp.display()
        
        # Create blank image for drawing.
        # Make sure to create image with mode '1' for 1-bit color.
        width = self._disp.width
        height = self._disp.height
        image = Image.new('1', (width, height))
        
        # Get drawing object to draw on image.
        draw = ImageDraw.Draw(image)
        
        # Draw a black filled box to clear the image.
        draw.rectangle((0, 0, width, height), outline=0, fill=0)

        # Load default font.
        font = ImageFont.load_default()

        # Retrieve values.
        timestr = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
        try:
            ssid = subprocess.check_output("/sbin/iwgetid -r", shell = True) \
                .decode("utf-8")
        except:
            ssid = "---"
        device_id = socket.gethostname()

        # Write text
        t, l = -2, 0
        draw.text((l, t + 0),  " "       + timestr,     font=font, fill=255)
        draw.text((l, t + 8),  "---------------------", font=font, fill=255)
        draw.text((l, t + 16), "SSID:  " + ssid,        font=font, fill=255)
        draw.text((l, t + 24), "DevID: " + device_id,   font=font, fill=255)

        # Display image.
        self._disp.image(image)
        self._disp.display()

        return [(self.sid(), int(time.time() * 1e9), None)]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="piot",
    version="0.0.3",
    author="Francisco Troncoso Pastoriza",
    author_email="frantp90@gmail.com",
    description="Load and run device drivers at specific intervals, mainly sensor reading",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/frantp/piot",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    keywords="driver sensor reading",
    install_requires=[
        "toml",
        "paho-mqtt",
        "filelock",
        "RPi.GPIO",
        "smbus2",
        "pyserial",
        "adafruit-circuitpython-adxl34x",
        "adafruit-circuitpython-amg88xx",
        "adafruit-circuitpython-bme280",
        "adafruit-circuitpython-bme680",
        "adafruit-circuitpython-dht",
        "adafruit-circuitpython-hcsr04",
        "adafruit-circuitpython-mlx90614",
        "adafruit-circuitpython-tsl2561",
        "adafruit-circuitpython-tsl2591",
        "Adafruit-SSD1306",
        "Pillow",
        "pypozyx-i2c",
    ],
    python_requires=">=3.5",
    entry_points={
        "console_scripts": [
            "piot = piot.core:main"
        ],
    },
)



from collections import OrderedDict
import time

from ..core import SerialDriver
from serial import SerialException


_REQUEST_SEQ = b"\x02\x01\x00\xFE"


def _checksum(res):
    return 0xFF - (sum(res[1:]) & 0xFF)


def _check(res):
    return _checksum(res[:-1]) != res[-1]


class Driver(SerialDriver):
    def __init__(self, port):
        super().__init__(port, 19200)


    def run(self):
        ts = int(time.time() * 1e9)
        res = self._cmd(_REQUEST_SEQ, 8)
        if res[0:3] != b"\x02\x10\x04" or _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        status, minutes, rint, rdec = res[3:7]
        return [(self.sid(), ts, OrderedDict([
            ("status", status),
            ("meastime", 60 * minutes),
            ("radon", (rint + rdec / 100) * 37),  # pCi/L -> Bq/m^3
        ]))]



from collections import OrderedDict
import time

from ..core import SMBusDriver


_CMD_RESET                    = 0xA0
_CMD_READ_STATUS              = 0xB0
_CMD_READ_MEASUREMENT_TIME    = 0xB1
_CMD_READ_MEASURING_VALUE_10M = 0xB2
_CMD_READ_MEASURING_VALUE_1M  = 0xB3
_CMD_READ_FIRMWARE_VERSION    = 0xB4


class Driver(SMBusDriver):
    def __init__(self, address=0x18, bus=1):
        super().__init__(bus)
        self._address = address


    def run(self):
        ts = int(time.time() * 1e9)
        status, vibration = self._bus.read_i2c_block_data(
            self._address, _CMD_READ_STATUS, 2)
        minutes, seconds = self._bus.read_i2c_block_data(
            self._address, _CMD_READ_MEASUREMENT_TIME, 2)
        gint10m, gdec10m = self._bus.read_i2c_block_data(
            self._address, _CMD_READ_MEASURING_VALUE_10M, 2)
        gint1m, gdec1m = self._bus.read_i2c_block_data(
            self._address, _CMD_READ_MEASURING_VALUE_1M, 2)
        firmm, firms = self._bus.read_i2c_block_data(
            self._address, _CMD_READ_FIRMWARE_VERSION, 2)
        return [(self.sid(), ts, OrderedDict([
            ("status", status),
            ("vibration", vibration),
            ("meastime", 60 * minutes + seconds),
            ("gamma10", gint10m + gdec10m / 100),
            ("gamma1", gint1m + gdec1m / 100),
            ("firmware", "{}.{}".format(firmm, firms)),
        ]))]



from collections import OrderedDict
import time

from ..core import I2CDriver
import board
import busio
from adafruit_mlx90614 import MLX90614


class Driver(I2CDriver):
    def __init__(self, address=0x5a):
        super().__init__()
        i2c = busio.I2C(board.SCL, board.SDA)
        self._sensor = MLX90614(i2c, address=address)


    def run(self):
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("ambient_temperature", self._sensor.ambient_temperature),
            ("object_temperature", self._sensor.object_temperature),
        ]))]



from collections import OrderedDict
import struct
import time

from ..core import SerialDriver
from serial import SerialException


def _checksum(res):
    return sum(res) & 0xFF


def _check(res):
    return _checksum(res[:-1]) != res[-1]


class Driver(SerialDriver):
    def __init__(self, port):
        super().__init__(port, 115200)


    def run(self):
        ts = int(time.time() * 1e9)
        self._serial.read_until(b"\x59\x59")
        res = b"\x59\x59" + self._serial.read(7)
        if _check(res):
            raise SerialException("Incorrect response: {}".format(res.hex()))
        distance, strength = struct.unpack("<HH", res[2:6])
        return [(self.sid(), ts, OrderedDict([
            ("distance", distance),
            ("strength", strength),
        ]))]



from collections import OrderedDict
import time

from ..core import DriverBase
import pypozyx


class Driver(DriverBase):
    def __init__(self, i2c=False, bus=1, port=None):
        super().__init__()
        self._sensor = pypozyx.PozyxI2C(bus) if i2c else \
            pypozyx.PozyxSerial(port or pypozyx.get_first_pozyx_serial_port())


    def close(self):
        if isinstance(self._sensor, pypozyx.PozyxI2C):
            self._sensor.bus.close()
        super().close()


    def run(self):
        position = pypozyx.Coordinates()
        while position.x == position.y == position.z == 0:
            self._sensor.doPositioning(position)
        return [(self.sid(), int(time.time() * 1e9), OrderedDict([
            ("pos_x", position.x),
            ("pos_y", position.y),
            ("pos_z", position.z),
        ]))]
